/* Copyright 2010 Jan-Mark Wams. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY JAN-MARK WAMS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL JAN-MARK WAMS OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * The views and conclusions contained in the software and
 * documentation are those of the authors and should not be
 * interpreted as representing official policies, either expressed or
 * implied, of Jan-Mark Wams.
 */

#include <stdio.h>
#include "trycatch.h"
DECLARE_TRYCATCH_GLOBALS;

/* Defined in trytest2.c. */
double f2(double);

typedef enum {EX_DIV_BY_ZERO = 1, EX_TOO_BIG} myex_t; 

double f(double x) 
{ 
  if (0 == x) raise(EX_DIV_BY_ZERO);
  if (x > 3.3) raise(EX_TOO_BIG);
  if (x == -1.1) raise(44);

  if (x == -2.0) 
    try               { printf("=== trying hard.\n"); 
                        raise(99); }
    catch(7)          { printf("=== seven\n"); }
    catch(EX_TOO_BIG) { printf("=== Ignoring EX_TOO_BIG.\n"); }
    catch(*)          { printf("=== Upagating exception %d.\n", excepno); 
                        upagate(*); /* No need for volatile. */}
    ensure            { printf("=== Ensuring all is ok.\n"); }

  return 1/x;
}

void prf(volatile double x)
{
  double y = -1234567890;

  try {
    printf("+++ Calling f(%e)\n", x);
    if (1.25 == x) y = f2(x); else y = f(x);
    printf("+++ No exception was raised.\n");
  }
  catch(EX_DIV_BY_ZERO) {
    printf("+++ Devide by zero! Retry with tiny value for x.\n");
    x = 1e-99;
    retry;        /* Note: argument x should be made volatile. */
  }
  catch(EX_TOO_BIG) {
    printf("+++ X is way too BIG!\n");
  }
  catch(*) {
    printf("+++ Unknown exception %d caught\n", excepno);
  }
  ensure {
    printf("+++ Ensuring y is positive.\n");
    if (y < 0) y *= -1.0;
  }
  
  printf("y == %e\n", y);
}

int main(int ac, char av) 
{
  try {
    printf("\nShould be Ok. No exceptions.\n");		       
    prf(1.1);  
    printf("\nZero.\n");					       
    prf(0.0);  
    printf("\nToo big.\n");					       
    prf(5.5);  
    printf("\nSpecial value to trigger unkown exception.\n");       
    prf(-1.1); 
    printf("\nSpecial value to trigger second level exceptions.\n");
    prf(-2.0); 
    printf("\nSpecial value to call trytest2.c:f2().\n");           
    prf(1.25); 
  }
  catch(*) {
    printf("*** Unknown exception %d caught in main().\n", excepno);
  }

  return 0;
}
